import asyncio
import inspect
import random
import time
from functools import wraps
from typing import Callable

from fastapi import Request
from loguru import logger


def retry(max_retries=3, delay=1, backoff=2, exceptions=(Exception,)):
    """
    A decorator for automatically retrying a function upon encountering specified exceptions.

    Args:
        max_retries (int): The maximum number of times to retry the function.
        delay (float): The initial delay between retries in seconds.
        backoff (float): The multiplier by which the delay should increase after each retry.
        exceptions (tuple): A tuple of exception classes upon which to retry.

    Returns:
        The return value of the wrapped function, if it succeeds.
        Raises the last encountered exception if the function never succeeds.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            retries = 0
            current_delay = delay
            while retries <= max_retries:
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    retries += 1
                    if retries == max_retries:
                        raise
                    logger.warning(
                        f"Error:{type(e)}\n"
                        f"Retrying `{func.__name__}` after {current_delay} seconds, retry : {retries}\n"
                    )
                    time.sleep(current_delay)
                    current_delay *= backoff

        return wrapper

    return decorator


def async_retry(
    max_retries=3,
    delay=1,
    backoff=2,
    exceptions=(Exception,),
    raise_callback_name=None,
    raise_handler_name=None,
):
    """
    An asynchronous decorator for automatically retrying an async function upon encountering specified exceptions.

    Args:
        max_retries (int): The maximum number of times to retry the function.
        delay (float): The initial delay between retries in seconds.
        backoff (float): The multiplier by which the delay should increase after each retry.
        exceptions (tuple): A tuple of exception classes upon which to retry.
        raise_callback_name (str): A callback function to call when the maximum number of retries is reached.
        raise_handler_name (str): A error handler function to call when the maximum number of retries is reached.

    Returns:
        The return value of the wrapped function, if it succeeds.
        Raises the last encountered exception if the function never succeeds.
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            retries = 0
            current_delay = delay
            while retries <= max_retries:
                try:
                    result = await func(*args, **kwargs)
                    return result
                except exceptions as e:

                    if retries == max_retries:
                        # Experimental
                        if raise_callback_name:
                            self = args[0]
                            callback: Callable = getattr(
                                self, raise_callback_name, None
                            )
                            if callback:
                                logger.warning(
                                    f"Calling raise callback {raise_callback_name}"
                                )
                                if getattr(self, 'client', None):
                                    await self.client.close()
                                callback()
                        if raise_handler_name:
                            self = args[0]
                            raise_handler: Callable = getattr(
                                self, raise_handler_name, None
                            )
                            if raise_handler:
                                logger.warning(
                                    f"Calling raise handler {raise_handler_name}"
                                )
                                raise_handler(e)
                        raise

                    retries += 1
                    logger.warning(
                        f"Error:{type(e)}\n"
                        f"Retrying `{func.__name__}` after {current_delay} seconds, retry : {retries}\n"
                    )
                    await asyncio.sleep(current_delay)
                    current_delay *= backoff

        return wrapper

    return decorator


def async_token_rate_limit(token_rate_limit: dict):
    """
    A decorator for rate-limiting requests based on tokens. It limits the rate at which tokens can be consumed
    for a particular route path.

    Args:
        token_rate_limit (dict): A dictionary mapping route paths to their respective token intervals (in seconds).

    Yields:
        value: The value from the wrapped asynchronous generator.

    Note:
        The 'request' object should be passed either as a keyword argument or as a positional argument to the
        decorated function.
    """

    def decorator(async_gen_func):
        @wraps(async_gen_func)
        async def wrapper(*args, **kwargs):
            request: Request = kwargs.get('request')
            if not request:
                # Try to find the request argument by position
                func_argspec = inspect.getfullargspec(async_gen_func)
                request_index = func_argspec.args.index('request')
                request = args[request_index]

            route_path = f"{request.scope.get('root_path')}{request.scope.get('path')}"
            token_interval = token_rate_limit.get(route_path, 0)

            async_gen = async_gen_func(*args, **kwargs)

            start_time = time.perf_counter()
            async for value in async_gen:
                if token_interval > 0:
                    current_time = time.perf_counter()
                    delta = current_time - start_time
                    delay = token_interval - delta
                    if delay > 0:
                        await asyncio.sleep(delay)
                    start_time = time.perf_counter()
                yield value

        return wrapper

    return decorator


def async_random_sleep(min_time=0, max_time=1):
    """
    Decorator that adds a random sleep time between min_time and max_time.

    Args:
        min_time (float, optional): The minimum sleep time in seconds.
        max_time (float, optional): The maximum sleep time in seconds.
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            if max_time == 0:
                return await func(*args, **kwargs)
            sleep_time = random.uniform(min_time, max_time)
            await asyncio.sleep(sleep_time)
            return await func(*args, **kwargs)

        return wrapper

    return decorator



from .base import OpenaiForward


def create_openai_proxies():
    """
    Generate OPENAI api style OpenaiForward objects.

    Returns:
        _objs (list): A list of OpenaiForward objects.
    """
    from ..settings import OPENAI_BASE_URL, OPENAI_ROUTE_PREFIX, PROXY

    _objs = []
    for base_url, route_prefix in zip(OPENAI_BASE_URL, OPENAI_ROUTE_PREFIX):
        _objs.append(OpenaiForward(base_url, route_prefix, PROXY))
    return _objs


openai_objs = create_openai_proxies()



from fastapi import FastAPI, Request, status
from fastapi.middleware.cors import CORSMiddleware
from slowapi import Limiter, _rate_limit_exceeded_handler
from slowapi.errors import RateLimitExceeded

from . import __version__, custom_slowapi
from .forward.extra import generic_objs
from .forward.openai import openai_objs
from .helper import normalize_route as normalize_route_path
from .settings import (
    BENCHMARK_MODE,
    RATE_LIMIT_BACKEND,
    RATE_LIMIT_STRATEGY,
    dynamic_request_rate_limit,
    get_limiter_key,
    show_startup,
)

limiter = Limiter(
    key_func=get_limiter_key,
    strategy=RATE_LIMIT_STRATEGY,
    storage_uri=RATE_LIMIT_BACKEND,
)
app = FastAPI(title="openai-forward", version=__version__)

app.state.limiter = limiter
app.add_exception_handler(RateLimitExceeded, _rate_limit_exceeded_handler)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.middleware("http")
async def normalize_route(request: Request, call_next):
    path = request.url.path
    request.scope["path"] = normalize_route_path(path)
    response = await call_next(request)
    return response


@app.get(
    "/healthz",
    summary="Perform a Health Check",
    response_description="Return HTTP Status Code 200 (OK)",
    status_code=status.HTTP_200_OK,
)
def healthz(request: Request):
    return "OK"


if BENCHMARK_MODE:
    from .cache.chat_completions import chat_completions_benchmark

    app.add_route(
        "/benchmark/v1/chat/completions",
        route=limiter.limit(dynamic_request_rate_limit)(chat_completions_benchmark),
        methods=["POST"],
    )


@app.on_event("startup")
async def startup():
    [await obj.build_client() for obj in openai_objs]
    [await obj.build_client() for obj in generic_objs]


@app.on_event("shutdown")
async def shutdown():
    [await obj.client.close() for obj in openai_objs]
    [await obj.client.close() for obj in generic_objs]


add_route = lambda obj: app.add_route(
    obj.ROUTE_PREFIX + "{api_path:path}",
    route=limiter.limit(dynamic_request_rate_limit)(obj.reverse_proxy),
    methods=["GET", "POST", "PUT", "DELETE", "OPTIONS", "HEAD", "PATCH", "TRACE"],
)
[add_route(obj) for obj in openai_objs]
[add_route(obj) for obj in generic_objs]

show_startup()



import inspect
from typing import List

from slowapi.extension import Limit, LimitGroup, RateLimitItem
from slowapi.wrappers import parse_many


def __iter__(self):
    if callable(self._LimitGroup__limit_provider):
        if (
            "key"
            in inspect.signature(self._LimitGroup__limit_provider).parameters.keys()
        ):
            assert (
                "request" in inspect.signature(self.key_function).parameters.keys()
            ), f"Limit provider function {self.key_function.__name__} needs a `request` argument"
            if self.request is None:
                raise Exception("`request` object can't be None")
            limit_raw = self._LimitGroup__limit_provider(
                self.key_function(self.request)
            )
        else:
            limit_raw = self._LimitGroup__limit_provider
    else:
        limit_raw = self._LimitGroup__limit_provider
    if limit_raw.lower().startswith("inf"):
        # limit_items = [RateLimitItem(0, 0, "second")]
        limit_items = ['-.-']
        self.exempt_when = lambda: True
    else:
        limit_items: List[RateLimitItem] = parse_many(limit_raw)
    for limit in limit_items:
        yield Limit(
            limit,
            self.key_function,
            self._LimitGroup__scope,
            self.per_method,
            self.methods,
            self.error_message,
            self.exempt_when,
            self.cost,
            self.override_defaults,
        )


LimitGroup.__iter__ = __iter__



from .base import GenericForward


def create_generic_proxies():
    """
    Generate extra forwarding objects.

    Returns:
        list: A list of GenericForward objects.
    """
    from ..settings import EXTRA_BASE_URL, EXTRA_ROUTE_PREFIX, PROXY

    _objs = []
    for base_url, route_prefix in zip(EXTRA_BASE_URL, EXTRA_ROUTE_PREFIX):
        _objs.append(GenericForward(base_url, route_prefix, PROXY))
    return _objs


generic_objs = create_generic_proxies()



import ast
import hashlib
import inspect
import os
import re
import time
from pathlib import Path
from typing import Dict, List, Union

import orjson
from fastapi import Request
from rich import print


def get_client_ip(request: Request):
    if "x-forwarded-for" in request.headers:
        return request.headers["x-forwarded-for"].split(",")[0]
    elif not request.client or not request.client.host:
        return "127.0.0.1"
    return request.client.host


def route_prefix_to_str(route_prefix):
    return route_prefix.replace('/', '_').strip("_") or "openai"


def get_unique_id():
    return hashlib.md5(str(time.time()).encode()).hexdigest()


def normalize_route(route):
    # If the route is only composed of slashes, return '/'
    if all(c == '/' for c in route):
        return '/'
    route = re.sub(
        r'//+', '/', route
    )  # Replace multiple consecutive '/' with a single '/'
    return route.rstrip('/')  # Remove trailing '/'


def relp(rel_path: Union[str, Path], parents=0, return_str=True, strict=False):
    currentframe = inspect.currentframe()
    f = currentframe.f_back
    for _ in range(parents):
        f = f.f_back
    current_path = Path(f.f_code.co_filename).parent
    pathlib_path = current_path / rel_path
    pathlib_path = pathlib_path.resolve(strict=strict)
    if return_str:
        return str(pathlib_path)
    else:
        return pathlib_path


def ls(_dir, *patterns, concat='extend', recursive=False):
    from glob import glob

    path_list = []
    for pattern in patterns:
        if concat == 'extend':
            path_list.extend(glob(os.path.join(_dir, pattern), recursive=recursive))
        else:
            path_list.append(glob(os.path.join(_dir, pattern), recursive=recursive))
    return path_list


def json_load(filepath: str, rel=False, mode="rb"):
    abs_path = relp(filepath, parents=1) if rel else filepath
    with open(abs_path, mode=mode) as f:
        return orjson.loads(f.read())


def json_dump(
    data: Union[List, Dict], filepath: str, rel=False, indent_2=False, mode="wb"
):
    orjson_option = 0
    if indent_2:
        orjson_option = orjson.OPT_INDENT_2
    abs_path = relp(filepath, parents=1) if rel else filepath
    with open(abs_path, mode=mode) as f:
        f.write(orjson.dumps(data, option=orjson_option))


def toml_load(filepath: str, rel=False):
    import toml

    abs_path = relp(filepath, parents=1) if rel else filepath
    return toml.load(abs_path)


def str2list(s: str, sep):
    if s:
        return [i.strip() for i in s.split(sep) if i.strip()]
    else:
        return []


def env2list(env_name: str, sep=","):
    return str2list(os.environ.get(env_name, "").strip(), sep=sep)


def env2dict(env_name: str) -> Dict:
    import json

    env_str = os.environ.get(env_name, "").strip()
    if not env_str:
        return {}
    return json.loads(env_str)


def format_route_prefix(route_prefix: str):
    if route_prefix:
        if route_prefix.endswith("/"):
            route_prefix = route_prefix[:-1]
        if not route_prefix.startswith("/"):
            route_prefix = "/" + route_prefix
    return route_prefix


def get_matches(messages: List[Dict], assistants: List[Dict]):
    msg_len, ass_len = len(messages), len(assistants)
    if msg_len != ass_len:
        print(f"Length mismatch between message({msg_len}) and assistant({ass_len}) ")

    cvt = lambda msg, ass: {
        "datetime": msg.get('datetime'),
        "ip": msg.get("ip") or msg.get('forwarded-for'),
        "model": msg.get("model"),
        "temperature": msg.get("temperature", 1),
        "messages": msg.get("messages"),
        "functions": msg.get("functions"),
        "is_function_call": ass.get("is_function_call"),
        "assistant": ass.get("assistant"),
    }

    msg_uid_dict = {m.pop("uid"): m for m in messages}
    ass_uid_dict = {a.pop("uid"): a for a in assistants}
    matches = [
        cvt(msg_uid_dict[uid], ass_uid_dict[uid])
        for uid in msg_uid_dict
        if uid in ass_uid_dict
    ]

    ref_len = max(msg_len, ass_len)
    if len(matches) != ref_len:
        print(f"There are {ref_len - len(matches)} mismatched items")
    return matches


def parse_log_to_list(log_path: str):
    def clean_str(original_str: str):
        try:
            return str(original_str).encode('utf-8', 'ignore').decode('utf-8')
        except Exception:
            print(f"{original_str=}")
            return ""

    with open(log_path, "r", encoding="utf-8") as f:
        messages, assistant = [], []
        for line in f.readlines():
            content: dict = ast.literal_eval(line)
            if content.get("messages"):
                clean_content = {}
                for key, value in content.items():
                    if key == "messages":
                        try:
                            clean_content[key] = [
                                {i['role']: clean_str(i['content'])} for i in value
                            ]
                        except Exception as e:
                            if not isinstance(value, list):
                                print(
                                    f"Warning(`message` should be a list): {content=}"
                                )
                            else:
                                print(f"Warning({e=}): {content=}")
                            continue
                    else:
                        clean_content[key] = value
                messages.append(clean_content)
            else:
                clean_content = {}
                for key, value in content.items():
                    if key == "assistant":
                        clean_content[key] = clean_str(value)
                    else:
                        clean_content[key] = value
                assistant.append(clean_content)
    return messages, assistant


def convert_chatlog_to_jsonl(log_path: str, target_path: str):
    """Convert single chat log to jsonl"""
    message_list, assistant_list = parse_log_to_list(log_path)
    content_list = get_matches(messages=message_list, assistants=assistant_list)
    json_dump(content_list, target_path, indent_2=True)


def get_log_files_from_folder(log_path: str):
    return ls(log_path, "*.log")


def convert_folder_to_jsonl(folder_path: str, target_path: str):
    """Convert chat log folder to jsonl"""
    log_files = get_log_files_from_folder(folder_path)
    messages = []
    assistants = []
    for log_path in log_files:
        msg, ass = parse_log_to_list(log_path)

        msg_len, ass_len = len(msg), len(ass)
        if msg_len != ass_len:
            print(
                f"{log_path}: Length mismatch between message({msg_len}) and assistant({ass_len}) "
            )
        messages.extend(msg)
        assistants.extend(ass)
    content_list = get_matches(messages=messages, assistants=assistants)
    json_dump(content_list, target_path, indent_2=True)
    print("Converted successfully")
    print(f"File saved to {target_path}")



import os
import platform

import fire
import uvicorn


class Cli:
    @staticmethod
    def run(port=8000, workers=1):
        """
        Runs the application using the Uvicorn server.

        Args:
            port (int): The port number on which to run the server. Default is 8000.
            workers (int): The number of worker processes to run. Default is 1.

        Returns:
            None
        """

        if platform.system() == "Windows":
            os.environ["TZ"] = ""

        ssl_keyfile = os.environ.get("ssl_keyfile", None) or None
        ssl_certfile = os.environ.get("ssl_certfile", None) or None
        uvicorn.run(
            app="openai_forward.app:app",
            host="0.0.0.0",
            port=port,
            workers=workers,
            app_dir="..",
            ssl_keyfile=ssl_keyfile,
            ssl_certfile=ssl_certfile,
        )

    @staticmethod
    def convert(log_folder: str = None, target_path: str = None):
        """
        Converts log files in a folder to a JSONL file.

        Args:
            log_folder (str, optional): The path to the folder containing the log files. Default is None.
            target_path (str, optional): The path to the target JSONL file. Default is None.

        Returns:
            None
        """
        from openai_forward.helper import convert_folder_to_jsonl, route_prefix_to_str
        from openai_forward.settings import OPENAI_ROUTE_PREFIX

        print(60 * '-')
        if log_folder is None:
            if target_path is not None:
                raise ValueError("target_path must be None when log_folder is None")
            _prefix_list = [route_prefix_to_str(i) for i in OPENAI_ROUTE_PREFIX]
            for prefix in _prefix_list:
                log_folder = f"./Log/{prefix}/chat"
                target_path = f"./Log/chat_{prefix}.json"
                print(f"Convert {log_folder}/*.log to {target_path}")
                convert_folder_to_jsonl(log_folder, target_path)
                print(60 * '-')
        else:
            print(f"Convert {log_folder}/*.log to {target_path}")
            convert_folder_to_jsonl(log_folder, target_path)
            print(60 * '-')


def main():
    fire.Fire(Cli)


if __name__ == "__main__":
    main()



from flaxkv import dbdict

from ..settings import CACHE_BACKEND, LOG_CACHE_DB_INFO

if CACHE_BACKEND.upper() == "LMDB":
    try:
        import lmdb
    except ImportError:
        raise ImportError("Please install LMDB: pip install lmdb")

    if LOG_CACHE_DB_INFO:
        db_dict = dbdict("./CACHE_LMDB", backend='lmdb', log="INFO", save_log=True)
    else:
        db_dict = dbdict("./CACHE_LMDB", backend='lmdb')

elif CACHE_BACKEND.upper() == "LEVELDB":
    try:
        import plyvel
    except ImportError:
        raise ImportError("Please install LevelDB: pip install plyvel")

    if LOG_CACHE_DB_INFO:
        db_dict = dbdict(
            "./CACHE_LEVELDB", backend='leveldb', log="INFO", save_log=True
        )
    else:
        db_dict = dbdict("./CACHE_LEVELDB", backend='leveldb')

elif CACHE_BACKEND.upper() == "MEMORY":
    db_dict = {}

else:
    raise ValueError(f"Unknown cache backend: {CACHE_BACKEND}")



import asyncio
import traceback
from asyncio import Queue
from itertools import cycle
from typing import Any, AsyncGenerator

import aiohttp
import anyio
from aiohttp import TCPConnector
from fastapi import HTTPException, Request, status
from flaxkv.pack import encode
from loguru import logger
from starlette.responses import BackgroundTask, Response, StreamingResponse

from ..cache.chat_completions import generate, stream_generate_efficient
from ..cache.database import db_dict
from ..content.openai import ChatLogger, CompletionLogger, WhisperLogger
from ..decorators import async_retry, async_token_rate_limit
from ..settings import *


class GenericForward:
    """
    Base class for handling request forwarding to another service.
    Provides methods for request validation, logging, and proxying.
    """

    validate_host = bool(IP_BLACKLIST or IP_WHITELIST)
    timeout = aiohttp.ClientTimeout(connect=TIMEOUT)

    def __init__(self, base_url: str, route_prefix: str, proxy=None):
        self.BASE_URL = base_url
        self.PROXY = proxy
        self.ROUTE_PREFIX = route_prefix
        self.client: aiohttp.ClientSession | None = None

    async def build_client(self):
        connector = TCPConnector(limit=500, limit_per_host=0, force_close=False)
        self.client = aiohttp.ClientSession(connector=connector, timeout=self.timeout)

    @staticmethod
    def validate_request_host(ip):
        """
        Validates the incoming IP address against a whitelist and blacklist.

        Args:
            ip (str): The IP address to validate.

        Raises:
            HTTPException: If the IP is not in the whitelist or if it is in the blacklist.
        """
        if (
            IP_WHITELIST
            and ip not in IP_WHITELIST
            or IP_BLACKLIST
            and ip in IP_BLACKLIST
        ):
            logger.warning(f"IP {ip} is unauthorized")
            raise HTTPException(
                status_code=status.HTTP_403_FORBIDDEN,
                detail=f"Forbidden Error",
            )

    @staticmethod
    @async_token_rate_limit(token_interval_conf)
    async def aiter_bytes(
        r: aiohttp.ClientResponse, request: Request
    ) -> AsyncGenerator[bytes, Any]:
        async for chunk, _ in r.content.iter_chunks():  # yield chunk one by one.
            yield chunk

    @async_retry(
        max_retries=3,
        delay=0.2,
        backoff=2,
        exceptions=(
            aiohttp.ServerTimeoutError,
            aiohttp.ServerConnectionError,
            aiohttp.ServerDisconnectedError,
            asyncio.TimeoutError,
            anyio.EndOfStream,
            RuntimeError,
        ),
        # raise_callback_name="build_client",
        raise_handler_name="handle_exception",
    )
    async def send(self, client_config: dict, data=None):

        return await self.client.request(
            method=client_config["method"],
            url=client_config['url'],
            data=data,
            headers=client_config["headers"],
            proxy=self.PROXY,
        )

    def handle_exception(self, e):
        if isinstance(
            e,
            (
                asyncio.TimeoutError,
                aiohttp.ServerConnectionError,
                aiohttp.ServerDisconnectedError,
                aiohttp.ServerTimeoutError,
            ),
        ):
            error_info = (
                f"{type(e)}: {e} | "
                f"Please check if your host can access [{self.BASE_URL}] successfully?"
            )
            status_code = status.HTTP_504_GATEWAY_TIMEOUT

        elif isinstance(e, anyio.EndOfStream):
            error_info = "EndOfStream Error: trying to read from a stream that has been closed from the other end."
            status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

        else:
            error_info = f"{type(e)}: {e}"
            status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

        logger.error(f"{error_info}\n{traceback.format_exc()}")
        raise HTTPException(status_code=status_code, detail=error_info)

    def prepare_client(self, request: Request, return_origin_header=False) -> dict:
        assert self.BASE_URL and self.ROUTE_PREFIX
        if self.validate_host:
            ip = get_client_ip(request)
            self.validate_request_host(ip)

        _url_path = f"{request.scope.get('root_path')}{request.scope.get('path')}"
        url_path = (
            _url_path[len(self.ROUTE_PREFIX) :]
            if self.ROUTE_PREFIX != '/'
            else _url_path
        )
        url = f"{self.BASE_URL}{url_path}?{request.url.query}"

        auth = request.headers.get("Authorization", "")

        if return_origin_header:
            headers = request.headers
        else:
            headers = {
                "content-type": request.headers.get("content-type", "application/json"),
                "authorization": auth,
            }
            for key, value in request.headers.items():
                if key.startswith("openai"):
                    headers[key] = value

        return {
            'auth': auth,
            'headers': headers,
            "method": request.method,
            'url': url,
            'url_path': url_path,
        }

    async def reverse_proxy(self, request: Request):
        assert self.client
        client_config = self.prepare_client(request, return_origin_header=True)

        r = await self.send(client_config, data=await request.body())

        return StreamingResponse(
            self.aiter_bytes(r, request),
            status_code=r.status,
            media_type=r.headers.get("content-type"),
            background=BackgroundTask(r.release),
        )


class OpenaiForward(GenericForward):
    """
    Derived class for handling request forwarding specifically for the OpenAI (Style) API.
    """

    _cycle_api_key = cycle(OPENAI_API_KEY)
    _no_auth_mode = OPENAI_API_KEY != [] and FWD_KEY == []

    def __init__(self, base_url: str, route_prefix: str, proxy=None):
        super().__init__(base_url, route_prefix, proxy)
        if LOG_CHAT or PRINT_CHAT:
            self.chat_logger = ChatLogger(self.ROUTE_PREFIX)
            self.completion_logger = CompletionLogger(self.ROUTE_PREFIX)
            self.whisper_logger = WhisperLogger(self.ROUTE_PREFIX)

    def _handle_result(
        self, buffer: bytearray, uid: str, route_path: str, request_method: str
    ):
        """
        Logs the result of the API call.

        Args:
            buffer (bytearray): List of bytes, usually from the API response.
            uid (str): Unique identifier for the request.
            route_path (str): API route path.
            request_method (str): HTTP method (e.g., 'GET', 'POST').

        Raises:
            Suppress all errors.
        """
        result_info = {}

        # If not configured to log or print chat, or the method is not POST, return early
        if not (LOG_CHAT or PRINT_CHAT) or request_method != "POST":
            return result_info

        try:
            # Determine which logger and method to use based on the route_path
            logger_instance = None
            if route_path == CHAT_COMPLETION_ROUTE:
                logger_instance = self.chat_logger
            elif route_path == COMPLETION_ROUTE:
                logger_instance = self.completion_logger
            elif route_path.startswith("/v1/audio/"):
                self.whisper_logger.log_buffer(buffer)
                return result_info

            # If a logger method is determined, parse bytearray and log if necessary
            if logger_instance:
                result_info = logger_instance.parse_bytearray(buffer)
                result_info["uid"] = uid

                if LOG_CHAT:
                    logger_instance.log(result_info)

                if PRINT_CHAT and logger_instance == self.chat_logger:
                    self.chat_logger.print_chat_info(result_info)

        except Exception:
            logger.warning(f"log result error:\n{traceback.format_exc()}")

        return result_info

    async def _handle_payload(self, request: Request, url_path: str):
        """
        Asynchronously logs the payload of the API call.

        Args:
            request (Request): The original FastAPI request object.
            url_path (str): The API route path.

        Returns:
            dict: A dictionary containing parsed messages, model, IP address, UID, and datetime.

        Raises:
            Suppress all errors.

        """
        payload_log_info = {"uid": None}

        if not (LOG_CHAT or PRINT_CHAT) or request.method != "POST":
            payload = await request.body()
            return False, payload_log_info, payload

        try:
            # Determine which logger and method to use based on the url_path
            logger_instance = None
            if url_path == CHAT_COMPLETION_ROUTE:
                logger_instance = self.chat_logger
            elif url_path == COMPLETION_ROUTE:
                logger_instance = self.completion_logger

            # If a logger method is determined, parse payload and log if necessary
            if logger_instance:
                payload_log_info, payload = await logger_instance.parse_payload(request)

                if payload_log_info and LOG_CHAT:
                    logger_instance.log(payload_log_info)

                if (
                    payload_log_info
                    and PRINT_CHAT
                    and logger_instance == self.chat_logger
                ):
                    self.chat_logger.print_chat_info(payload_log_info)
            else:
                payload = await request.body()

        except Exception as e:
            logger.warning(
                f"log chat error:\nhost:{request.client.host} method:{request.method}: {traceback.format_exc()}"
            )
            payload = await request.body()

        valid = True if payload_log_info['uid'] is not None else False
        return valid, payload_log_info, payload

    @staticmethod
    async def read_chunks(r: aiohttp.ClientResponse, queue):
        buffer = bytearray()

        # Efficiency Mode
        if ITER_CHUNK_TYPE == "efficiency":
            # yield all available data as soon as it is received.
            async for chunk in r.content.iter_any():
                buffer.extend(chunk)
                await queue.put(chunk)

        # Precision Mode
        else:
            async for chunk, _ in r.content.iter_chunks():
                buffer.extend(chunk)
                await queue.put(chunk)

        await queue.put(buffer)  # add all information when the stream ends

    @async_token_rate_limit(token_interval_conf)
    async def aiter_bytes(
        self,
        r: aiohttp.ClientResponse,
        request: Request,
        route_path: str,
        uid: str,
        cache_key: str | None = None,
    ):
        """
        Asynchronously iterates through the bytes in the given aiohttp.ClientResponse object
        and yields each chunk while also logging the request and response data.

        Args:
            r (aiohttp.ClientResponse): The aiohttp.ClientResponse object.
            request (Request): The original FastAPI request object.
            route_path (str): The API route path.
            uid (str): Unique identifier for the request.
            cache_key (bytes): The cache key.

        Returns:
             AsyncGenerator[bytes]: Each chunk of bytes from the server's response.
        """

        queue_is_complete = False

        queue = Queue()
        # todo:
        task = asyncio.create_task(self.read_chunks(r, queue))
        try:
            while True:
                chunk = await queue.get()
                if not isinstance(chunk, bytes):
                    queue.task_done()
                    queue_is_complete = True
                    break
                yield chunk
        except Exception:
            logger.warning(
                f"aiter_bytes error:\nhost:{request.client.host} method:{request.method}: {traceback.format_exc()}"
            )
        finally:
            if not task.done():
                task.cancel()
            r.release()

        if uid:
            if r.ok and queue_is_complete:
                target_info = self._handle_result(
                    chunk, uid, route_path, request.method
                )
                print(f"{target_info=}")
                if target_info and CACHE_CHAT_COMPLETION and cache_key is not None:
                    cached_value = db_dict.get(cache_key, [])
                    cached_value.append(target_info["assistant"])
                    db_dict[cache_key] = cached_value
            elif chunk is not None:
                logger.warning(
                    f'uid: {uid}\n status: {r.status}\n {chunk.decode("utf-8")}'
                )
            else:
                logger.warning(f'uid: {uid}\n' f'{r.status}')

    def handle_authorization(self, client_config):
        auth, auth_prefix = client_config["auth"], "Bearer "
        if self._no_auth_mode or auth and auth[len(auth_prefix) :] in FWD_KEY:
            auth = auth_prefix + next(self._cycle_api_key)
            client_config["headers"]["Authorization"] = auth
        return auth

    @staticmethod
    def _get_cached_response(payload_info, valid_payload, request):
        """
        Attempts to retrieve a cached response based on the current request's payload information.

        This function constructs a cache key based on various aspects of the request payload,
        checks if the response for this key has been cached, and if so, constructs and returns
        the appropriate cached response.

        Returns:
            Tuple[Union[Response, None], Union[str, None]]:
                - Response (Union[Response, None]): The cached response if available; otherwise, None.
                - cache_key (Union[str, None]): The constructed cache key for the request. None if caching is not applicable.

        Note:
            If a cache hit occurs, the cached response is immediately returned without contacting the external server.
        """
        # todo: refactor this function

        def construct_cache_key():
            elements = [
                payload_info["n"],
                payload_info['messages'],
                payload_info['model'],
                payload_info["max_tokens"],
                payload_info['response_format'],
                payload_info['seed'],
                # payload_info['temperature'],
                payload_info["tools"],
                payload_info["tool_choice"],
            ]

            return encode(elements)

        def get_response_from_cache(key):
            logger.info(f'uid: {payload_info["uid"]} >>>>> [cache hit]')
            cache_values = db_dict[key]
            # todo: handle multiple choices
            cache_value = cache_values[-1]
            if isinstance(cache_value, list):
                text = None
                tool_calls = cache_value
            else:
                text = cache_value
                tool_calls = None

            if payload_info["stream"]:
                return StreamingResponse(
                    stream_generate_efficient(
                        payload_info['model'],
                        text,
                        tool_calls,
                        request,
                    ),
                    status_code=200,
                    media_type="text/event-stream",
                )

            else:
                usage = {
                    "prompt_tokens": 0,
                    "completion_tokens": 0,
                    "total_tokens": 0,
                }
                return Response(
                    content=generate(payload_info['model'], text, tool_calls, usage),
                    media_type="application/json",
                )

        if not (CACHE_CHAT_COMPLETION and valid_payload):
            return None, None

        cache_key = construct_cache_key()

        if payload_info['caching'] and cache_key in db_dict:
            return get_response_from_cache(cache_key), cache_key

        return None, cache_key

    async def reverse_proxy(self, request: Request):
        """
        Asynchronously handles reverse proxying the incoming request.

        Args:
            request (Request): The incoming FastAPI request object.

        Returns:
            StreamingResponse: A FastAPI StreamingResponse containing the server's response.
        """
        client_config = self.prepare_client(request, return_origin_header=False)
        url_path = client_config["url_path"]

        self.handle_authorization(client_config)
        valid_payload, payload_info, payload = await self._handle_payload(
            request, url_path
        )
        uid = payload_info["uid"]

        cached_response, cache_key = self._get_cached_response(
            payload_info, valid_payload, request
        )
        if cached_response:
            return cached_response

        r = await self.send(client_config, data=payload)
        return StreamingResponse(
            self.aiter_bytes(r, request, url_path, uid, cache_key),
            status_code=r.status,
            media_type=r.headers.get("content-type"),
        )



import time
from typing import List

import orjson
from fastapi import Request
from loguru import logger
from orjson import JSONDecodeError

from ..helper import get_client_ip, get_unique_id, route_prefix_to_str
from .helper import markdown_print, parse_sse_buffer, print


class CompletionLogger:
    def __init__(self, route_prefix: str):
        """
        Initialize the Completions logger with a route prefix.

        Args:
            route_prefix (str): The prefix used for routing, e.g., '/openai'.
        """
        _prefix = route_prefix_to_str(route_prefix)
        kwargs = {_prefix + "_completion": True}
        self.logger = logger.bind(**kwargs)

    @staticmethod
    async def parse_payload(request: Request):
        uid = get_unique_id()
        payload = await request.json()
        print(f"{payload=}")

        content = {
            "prompt": payload['prompt'],
            "model": payload['model'],
            "stream": payload.get("stream", True),
            "temperature": payload.get("temperature", 1),
            "ip": get_client_ip(request) or "",
            "uid": uid,
            "datetime": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()),
        }
        return content, payload

    @staticmethod
    def parse_bytearray(buffer: bytearray):
        """
        Parses a bytearray, usually from an API response, into a dictionary containing various information.

        Args:
            buffer (List[bytes]): A list of bytes to parse.

        Returns:
            Dict[str, Any]: A dictionary containing metadata and content. The keys include:
                - "text" (str)
        """
        start_token = "data: "
        start_token_len = len(start_token)
        if buffer.startswith(b'data: '):
            txt_lines = parse_sse_buffer(buffer)
            stream = True
            first_dict = orjson.loads(txt_lines[0][start_token_len:])
            text = first_dict["choices"][0]["text"]
        else:
            stream = False
            first_dict = orjson.loads(buffer)
            text = first_dict["choices"][0]["text"]

        target_info = dict()
        verbose = False
        if verbose:
            target_info["created"] = first_dict["created"]
            target_info["id"] = first_dict["id"]
            target_info["model"] = first_dict["model"]
            target_info["usage"] = first_dict.get("usage")

        target_info['text'] = text

        if not stream:
            return target_info

        # loop for stream
        for line in txt_lines[1:]:
            if line.startswith(start_token):
                try:
                    line_dict = orjson.loads(line[start_token_len:])
                    text += line_dict["choices"][0]["text"]
                except (JSONDecodeError, KeyError):
                    pass
        target_info['text'] = text
        return target_info

    def log(self, chat_info: dict):
        """
        Log chat information to the logger bound to this instance.

        Args:
            chat_info (dict): A dictionary containing chat information to be logged.
        """
        self.logger.debug(f"{chat_info}")


class ChatLogger:
    def __init__(self, route_prefix: str):
        """
        Initialize the Chat Completions logger with a route prefix.

        Args:
            route_prefix (str): The prefix used for routing, e.g., '/openai'.
        """
        _prefix = route_prefix_to_str(route_prefix)
        kwargs = {_prefix + "_chat": True}
        self.logger = logger.bind(**kwargs)

    @staticmethod
    async def parse_payload(request: Request):
        """
        Asynchronously parse the payload from a FastAPI request.

        Args:
            request (Request): A FastAPI request object.

        Returns:
            dict: A dictionary containing parsed messages, model, IP address, UID, and datetime.
        """
        uid = get_unique_id()
        payload = await request.json()

        # functions = payload.get("functions") # deprecated
        # if functions:
        #     info = {
        #         "functions": functions, # Deprecated in favor of `tools`
        #         "function_call": payload.get("function_call", None), # Deprecated in favor of `tool_choice`
        #     }
        info = {}
        info.update(
            {
                "messages": payload["messages"],
                "model": payload["model"],
                "stream": payload.get("stream", False),
                "max_tokens": payload.get("max_tokens", None),
                "response_format": payload.get("response_format", None),
                "n": payload.get("n", 1),
                "temperature": payload.get("temperature", 1),
                "top_p": payload.get("top_p", 1),
                "logit_bias": payload.get("logit_bias", None),
                "frequency_penalty": payload.get("frequency_penalty", 0),
                "presence_penalty": payload.get("presence_penalty", 0),
                "seed": payload.get("seed", None),
                "stop": payload.get("stop", None),
                "user": payload.get("user", None),
                "tools": payload.get("tools", None),
                "tool_choice": payload.get("tool_choice", None),
                "ip": get_client_ip(request) or "",
                "uid": uid,
                "caching": payload.pop("caching", False),  # pop caching
                "datetime": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()),
            }
        )
        return info, orjson.dumps(payload)

    def parse_bytearray(self, buffer: bytearray):
        """
        Parses a bytearray, usually from an API response, into a dictionary containing various information.

        Args:
            buffer (List[bytes]): A list of bytes to parse.

        Returns:
            Dict[str, Any]: A dictionary containing metadata and content. The keys include:
                - "assistant" (str): content
                - "is_function_call" (boolean)
        """
        start_token = "data: "
        start_token_len = len(start_token)
        if buffer.startswith(b'data: '):
            txt_lines = parse_sse_buffer(buffer)
            stream = True
            first_dict = orjson.loads(txt_lines[0][start_token_len:])
            # todo: multiple choices
            msg = first_dict["choices"][0]["delta"]
        else:
            stream = False
            first_dict = orjson.loads(buffer)
            # todo: multiple choices
            msg = first_dict["choices"][0]["message"]

        target_info = dict()
        verbose = False
        if verbose:
            target_info["created"] = first_dict["created"]
            target_info["id"] = first_dict["id"]
            target_info["model"] = first_dict["model"]
            target_info["role"] = msg["role"]
        role = msg["role"]  # always be "assistant"

        content, tool_calls = msg.get("content"), msg.get("tool_calls")
        if tool_calls:
            """
            tool_calls:
                [{
                "index": 0,
                "id": 'xx',
                'type": 'function',
                'function': {'name': 'xxx', 'arguments': ''}
                 }]
            """
            target_info[role] = tool_calls
            target_info["is_tool_calls"] = True
            parse_content_key = "tool_calls"

        else:
            target_info[role] = content
            target_info["is_tool_calls"] = False
            parse_content_key = "content"

        if not stream:
            return target_info

        # loop for stream
        stream_content = ""
        for line in txt_lines[1:]:
            if line.startswith(start_token):
                stream_content += self._parse_one_line_content(
                    line[start_token_len:], parse_content_key
                )
        if target_info['is_tool_calls']:
            tool_calls[0]['function']['arguments'] = stream_content
            target_info[role] = tool_calls
        else:
            target_info[role] = stream_content
        return target_info

    @staticmethod
    def _parse_one_line_content(line: str, parse_key="content"):
        """
        Helper method to parse a single line.

        Args:
            line (str): The line to parse.
            parse_key (str): .

        Returns:
            str: The parsed content from the line.
        """
        try:
            line_dict = orjson.loads(line)
            if parse_key == "content":
                return line_dict["choices"][0]["delta"][parse_key]
            elif parse_key == "tool_calls":
                tool_calls = line_dict["choices"][0]["delta"]["tool_calls"]
                return tool_calls[0]["function"]['arguments']
            else:
                logger.error(f"Unknown parse key: {parse_key}")
                return ""
        except JSONDecodeError:
            return ""
        except KeyError:
            return ""

    def log(self, chat_info: dict):
        """
        Log chat information to the logger bound to this instance.

        Args:
            chat_info (dict): A dictionary containing chat information to be logged.
        """
        self.logger.debug(f"{chat_info}")

    @staticmethod
    def print_chat_info(chat_info: dict):
        """
        Print chat information in a formatted manner.

        Args:
            chat_info (dict): A dictionary containing chat information to be printed.
        """
        messages = chat_info.get("messages")
        if messages:
            print(50 * "-", role='other')
            for msg in messages:
                for key, value in msg.items():
                    markdown_print(f"{key}: {value}", role=key)
            print(
                f"{chat_info.get('ip')}@{chat_info.get('model')} uid: {chat_info.get('uid')}",
                role='other',
            )
            print(50 * "-", role='other')
        else:
            assistant = chat_info.get("assistant")
            if assistant:
                print(77 * "=", role='assistant')
                markdown_print(f"assistant: {assistant}", role='assistant')
                print(f'uid: {chat_info.get("uid")}')
                print(77 * "=", role='assistant')


class WhisperLogger:
    def __init__(self, route_prefix: str):
        """
        Initialize the Audio logger with a route prefix.

        Args:
            route_prefix (str): The prefix used for routing, e.g., '/openai'.
        """
        _prefix = route_prefix_to_str(route_prefix)
        self.logger = logger.bind(**{f"{_prefix}_whisper": True})

    def log_buffer(self, bytes_: bytes):
        """
        Add a log entry containing the decoded text from bytes.

        Args:
            bytes_ (bytes): The byte string to decode and log.
        """
        text_content = bytes_.decode("utf-8")
        self.logger.info(text_content)



import functools
import logging
import os
import sys
import time
from datetime import datetime

import pytz
from loguru import logger

from ..helper import route_prefix_to_str


class InterceptHandler(logging.Handler):
    def emit(self, record):
        # Get corresponding Loguru level if it exists
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message
        frame, depth = logging.currentframe(), 6
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1
        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )


def setting_log(
    save_file=False,
    openai_route_prefix=None,
    log_name="openai_forward",
    multi_process=True,
    **kwargs,
):
    """
    Configures the logging settings for the application.
    """

    tz = os.environ.get("TZ", "").strip()
    if tz and hasattr(time, "tzset"):

        def get_utc_offset(timezone_str):
            timezone = pytz.timezone(timezone_str)
            offset_seconds = timezone.utcoffset(datetime.now()).total_seconds()
            offset_hours = offset_seconds // 3600
            return f"UTC{-int(offset_hours):+d}"

        try:
            os.environ["TZ"] = get_utc_offset(tz)
            time.tzset()
        except Exception:
            pass

    logging.root.handlers = [InterceptHandler()]
    for name in logging.root.manager.loggerDict.keys():
        logging.getLogger(name).handlers = []
        logging.getLogger(name).propagate = True

    print_chat = kwargs.get('print_chat')
    config_handlers = [
        {"sink": sys.stdout, "level": "INFO" if print_chat else "DEBUG"},
    ]

    for prefix in openai_route_prefix or []:

        _prefix = route_prefix_to_str(prefix)

        config_handlers.extend(
            [
                {
                    "sink": f"./Log/{_prefix}/chat/chat.log",
                    "enqueue": multi_process,
                    "rotation": "50 MB",
                    "filter": lambda record: f"{_prefix}_chat" in record["extra"],
                    "format": "{message}",
                },
                {
                    "sink": f"./Log/{_prefix}/completion/completion.log",
                    "enqueue": multi_process,
                    "rotation": "50 MB",
                    "filter": lambda record: f"{_prefix}_completion" in record["extra"],
                    "format": "{message}",
                },
                {
                    "sink": f"./Log/{_prefix}/whisper/whisper.log",
                    "enqueue": multi_process,
                    "rotation": "30 MB",
                    "filter": lambda record: f"{_prefix}_whisper" in record["extra"],
                    "format": "{message}",
                },
            ]
        )

    if save_file:
        config_handlers += [
            {
                "sink": f"./Log/{log_name}.log",
                "enqueue": multi_process,
                "rotation": "100 MB",
                "level": "INFO",
            }
        ]

    logger_config = {"handlers": config_handlers}
    logger.configure(**logger_config)



import itertools
import os

import limits
from fastapi import Request

from .console import print_rate_limit_info, print_startup_info
from .content.config import setting_log
from .helper import env2dict, env2list, format_route_prefix, get_client_ip

additional_start_info = {}

TIMEOUT = float(os.environ.get("TIMEOUT", "").strip() or "10")

ITER_CHUNK_TYPE = (
    os.environ.get("ITER_CHUNK_TYPE", "").strip() or "efficiency"
)  # Options: efficiency, precision

CHAT_COMPLETION_ROUTE = (
    os.environ.get("CHAT_COMPLETION_ROUTE", "").strip() or "/v1/chat/completions"
)

COMPLETION_ROUTE = os.environ.get("COMPLETION_ROUTE", "").strip() or "/v1/completions"

ENV_VAR_SEP = ","
OPENAI_BASE_URL = env2list("OPENAI_BASE_URL", sep=ENV_VAR_SEP) or [
    "https://api.openai.com"
]

OPENAI_ROUTE_PREFIX = [
    format_route_prefix(i) for i in env2list("OPENAI_ROUTE_PREFIX", sep=ENV_VAR_SEP)
] or ['/']

EXTRA_BASE_URL = env2list("EXTRA_BASE_URL", sep=ENV_VAR_SEP)
EXTRA_ROUTE_PREFIX = [
    format_route_prefix(i) for i in env2list("EXTRA_ROUTE_PREFIX", sep=ENV_VAR_SEP)
]

BENCHMARK_MODE = os.environ.get("BENCHMARK_MODE", "false").strip().lower() == "true"
if BENCHMARK_MODE:
    additional_start_info["benchmark_mode"] = BENCHMARK_MODE

LOG_CHAT = os.environ.get("LOG_CHAT", "False").strip().lower() == "true"
PRINT_CHAT = os.environ.get("PRINT_CHAT", "False").strip().lower() == "true"
if LOG_CHAT:
    setting_log(openai_route_prefix=OPENAI_ROUTE_PREFIX, print_chat=PRINT_CHAT)
    additional_start_info["log_chat"] = LOG_CHAT

if PRINT_CHAT:
    additional_start_info["print_chat"] = True

CACHE_CHAT_COMPLETION = (
    os.environ.get("CACHE_CHAT_COMPLETION", "false").strip().lower() == "true"
)
LOG_CACHE_DB_INFO = (
    os.environ.get("LOG_CACHE_DB_INFO", "false").strip().lower() == "true"
)
CACHE_BACKEND = os.environ.get("CACHE_BACKEND", "MEMORY").strip()
if CACHE_CHAT_COMPLETION:
    additional_start_info["cache_backend"] = CACHE_BACKEND

IP_WHITELIST = env2list("IP_WHITELIST", sep=ENV_VAR_SEP)
IP_BLACKLIST = env2list("IP_BLACKLIST", sep=ENV_VAR_SEP)

OPENAI_API_KEY = env2list("OPENAI_API_KEY", sep=ENV_VAR_SEP)
FWD_KEY = env2list("FORWARD_KEY", sep=ENV_VAR_SEP)

PROXY = os.environ.get("PROXY", "").strip() or None

if PROXY:
    additional_start_info["proxy"] = PROXY

GLOBAL_RATE_LIMIT = os.environ.get("GLOBAL_RATE_LIMIT", "").strip() or "inf"
RATE_LIMIT_BACKEND = os.environ.get("REQ_RATE_LIMIT_BACKEND", "").strip() or None
RATE_LIMIT_STRATEGY = (
    os.environ.get("RATE_LIMIT_STRATEGY", "fixed-window").strip() or "fixed-window"
)
req_rate_limit_dict = env2dict('REQ_RATE_LIMIT')


def get_limiter_key(request: Request):
    limiter_prefix = f"{request.scope.get('root_path')}{request.scope.get('path')}"
    key = f"{limiter_prefix}-{get_client_ip(request)}"
    return key


def dynamic_request_rate_limit(key: str):
    for route in req_rate_limit_dict:
        if key.startswith(route):
            return req_rate_limit_dict[route]
    return GLOBAL_RATE_LIMIT


def cvt_token_rate_to_interval(token_rate_limit: str):
    if token_rate_limit:
        rate_limit_item = limits.parse(token_rate_limit)
        token_interval = (
            rate_limit_item.multiples * rate_limit_item.GRANULARITY.seconds
        ) / rate_limit_item.amount
    else:
        token_interval = 0
    return token_interval


token_rate_limit_conf = env2dict("TOKEN_RATE_LIMIT")
token_interval_conf = {}
for route, rate_limit in token_rate_limit_conf.items():
    token_interval_conf[route] = cvt_token_rate_to_interval(rate_limit)

styles = itertools.cycle(
    ["#7CD9FF", "#BDADFF", "#9EFFE3", "#f1b8e4", "#F5A88E", "#BBCA89"]
)


def show_startup():
    for base_url, route_prefix in zip(OPENAI_BASE_URL, OPENAI_ROUTE_PREFIX):
        print_startup_info(
            base_url,
            route_prefix,
            OPENAI_API_KEY,
            FWD_KEY,
            style=next(styles),
            **additional_start_info,
        )
    for base_url, route_prefix in zip(EXTRA_BASE_URL, EXTRA_ROUTE_PREFIX):
        extra_additional_start_info = {}
        print_startup_info(
            base_url,
            route_prefix,
            "",
            "",
            style=next(styles),
            **extra_additional_start_info,
        )

    print_rate_limit_info(
        RATE_LIMIT_BACKEND,
        RATE_LIMIT_STRATEGY,
        GLOBAL_RATE_LIMIT,
        req_rate_limit_dict,
        token_rate_limit_conf,
    )
